/*
 * MPI sync clocks
 * Copyright (C) 2006-2020 (see AUTHORS file)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 */

/** @file measure drift of multiple synchronized clocks mechanisms
 */

#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <mpi.h>
#include <mpi_sync_clocks.h>

#define ROUNDS_DEFAULT 100
#define ROUNDS_TIME_SECONDS 5
#define ROUNDTRIPS 1000

static int rank = -1, worldsize = -1;

struct nm_clock_method_s
{
  const char*name;
  /** init the clock method & set time origin */
  void (*init)(void);
  /** get local time in usec. */
  double (*get_time_usec)(void);
  void (*finalize)(void);
  /** translate local time to global time */
  double (*translate_time)(double t, int n);
  struct nm_clock_drift_s*p_drift;
};

/* ** generic synchronization ****************************** */

static void generic_sync(int dest, double(*get_time)(void), double*t_master, double*t_slave)
{
  const int tag_sync = 42;
  if(rank == 0)
    {
      /* pingpong to get a common event on both master time and slave time */
      double min_lat = -1.0;
      double t_begin = -1.0, t_mid = -1.0, t_end = -1.0;
      int j;
      for(j = 0; j < ROUNDTRIPS; j++)
        {
          double t_dummy = -1.0;
          t_begin = (*get_time)();
          MPI_Send(&t_dummy, 1, MPI_DOUBLE, dest, tag_sync, MPI_COMM_WORLD);
          MPI_Recv(&t_mid, 1, MPI_DOUBLE, dest, tag_sync, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
          t_end = (*get_time)();
          const double lat = (t_end - t_begin) / 2.0;
          if(min_lat < 0 || lat < min_lat)
            {
              min_lat = lat;
            }
        }
      *t_master = t_begin;
      *t_slave  = t_mid - min_lat;
      MPI_Send(t_master, 1, MPI_DOUBLE, dest, tag_sync, MPI_COMM_WORLD);
      MPI_Send(t_slave, 1, MPI_DOUBLE, dest, tag_sync, MPI_COMM_WORLD);
    }
  else
    {
      int j;
      for(j = 0; j < ROUNDTRIPS; j++)
        {
          double t_dummy = -1.0;
          MPI_Recv(&t_dummy, 1, MPI_DOUBLE, dest, tag_sync, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
          double t_mid = (*get_time)();
          MPI_Send(&t_mid, 1, MPI_DOUBLE, dest, tag_sync, MPI_COMM_WORLD);
        }
      MPI_Recv(t_master, 1, MPI_DOUBLE, dest, tag_sync, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
      MPI_Recv(t_slave, 1, MPI_DOUBLE, dest, tag_sync, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
    }
}

/* ** MPI sync clocks ************************************** */

static mpi_sync_clocks_t p_mpi_sync_clocks = NULL;

static void sync_clocks_init(void)
{
  p_mpi_sync_clocks = mpi_sync_clocks_init(MPI_COMM_WORLD);
}

static void sync_clocks_finalize(void)
{
  /* final sync */
  mpi_sync_clocks_synchronize(p_mpi_sync_clocks);
}

static double sync_clocks_get_time_usec(void)
{
  const double t = mpi_sync_clocks_get_time_usec(p_mpi_sync_clocks);
  return t;
}

static double sync_clocks_translate_time(double t_slave, int n)
{
  return mpi_sync_clocks_remote_to_global(p_mpi_sync_clocks, n, t_slave);
}

static struct nm_clock_method_s sync_clocks_method =
  {
   .name           = "mpi_sync_clocks",
   .init           = &sync_clocks_init,
   .get_time_usec  = &sync_clocks_get_time_usec,
   .finalize       = &sync_clocks_finalize,
   .translate_time = &sync_clocks_translate_time
  };


/* ** offset clock ***************************************** */

static struct
{
  struct timespec orig; /**< time origin for local time */
  double*offsets;       /**< offset of nodes with node #0 (allocated only on node #0) */
} offset_clock;

static void offset_clock_init(void);
static double offset_clock_get_time_usec(void);
static double offset_clock_translate_time(double t, int n);

static void offset_clock_init(void)
{
  offset_clock.offsets = NULL;
  clock_gettime(CLOCK_MONOTONIC, &offset_clock.orig);
  if(rank == 0)
    {
      offset_clock.offsets = malloc(sizeof(double) * worldsize);
      int i;
      for(i = 0; i < worldsize; i++)
        {
          double t_master = -1.0, t_slave = -1.0;
          generic_sync(i, &offset_clock_get_time_usec, &t_master, &t_slave);
          offset_clock.offsets[i] = t_master - t_slave;
        }
    }
  else
    {
      double t_master = -1.0, t_slave = -1.0;
      generic_sync(0, &offset_clock_get_time_usec, &t_master, &t_slave);
    }
}

static double offset_clock_get_time_usec(void)
{
  struct timespec t;
  clock_gettime(CLOCK_MONOTONIC, &t);
  struct timespec diff;
  if(t.tv_nsec > offset_clock.orig.tv_nsec)
    {
      diff.tv_sec  = t.tv_sec - offset_clock.orig.tv_sec;
      diff.tv_nsec = t.tv_nsec - offset_clock.orig.tv_nsec;
    }
  else
    {
      diff.tv_sec  = t.tv_sec - offset_clock.orig.tv_sec - 1;
      diff.tv_nsec = t.tv_nsec - offset_clock.orig.tv_nsec + 1000000000L;
    }
  double delay = 1000000.0 * diff.tv_sec + diff.tv_nsec / 1000.0;
  return delay;
}

static double offset_clock_translate_time(double t_slave, int n)
{
  return t_slave + offset_clock.offsets[n];
}

static struct nm_clock_method_s offset_clock_method =
  {
   .name           = "offset_clock",
   .init           = &offset_clock_init,
   .get_time_usec  = &offset_clock_get_time_usec,
   .finalize       = NULL,
   .translate_time = &offset_clock_translate_time
  };


/* ** extrapolate clock ************************************ */

static struct
{
  struct timespec orig; /**< local origin */
  struct extrapolate_clock_node_s
  {
    double t_orig; /**< synchronization origin, to compute skew */
    double offset; /**< t_slave + offset = t_master */
    double skew;   /**< t_slave * skew = t_master */
  }*nodes;
} extrapolate_clock;

static void extrapolate_clock_init(void);
static double extrapolate_clock_get_time_usec(void);
static double extrapolate_clock_translate_time(double t, int n);

static void extrapolate_clock_init(void)
{
  extrapolate_clock.nodes = NULL;
  clock_gettime(CLOCK_MONOTONIC_RAW, &extrapolate_clock.orig);
  if(rank == 0)
    {
      extrapolate_clock.nodes = malloc(sizeof(struct extrapolate_clock_node_s) * worldsize);
      int i;
      for(i = 0; i < worldsize; i++)
        {
          double t_master = -1.0, t_slave = -1.0;
          generic_sync(i, &extrapolate_clock_get_time_usec, &t_master, &t_slave);
          extrapolate_clock.nodes[i].offset = t_master - t_slave;
          extrapolate_clock.nodes[i].t_orig = t_master;
        }
      sleep(10);
      for(i = 0; i < worldsize; i++)
        {
          double t_master = -1.0, t_slave = -1.0;
          generic_sync(i, &extrapolate_clock_get_time_usec, &t_master, &t_slave);
          const double t_diff_master = t_master - extrapolate_clock.nodes[i].t_orig;
          const double t_diff_slave  = t_slave  - (extrapolate_clock.nodes[i].t_orig - extrapolate_clock.nodes[i].offset);
          extrapolate_clock.nodes[i].skew = t_diff_master / t_diff_slave;
        }
    }
  else
    {
      double t_master = -1.0, t_slave = -1.0;
      generic_sync(0, &extrapolate_clock_get_time_usec, &t_master, &t_slave);
      generic_sync(0, &extrapolate_clock_get_time_usec, &t_master, &t_slave);
    }
}

static double extrapolate_clock_get_time_usec(void)
{
  struct timespec t;
  clock_gettime(CLOCK_MONOTONIC_RAW, &t);
  struct timespec diff;
  if(t.tv_nsec > extrapolate_clock.orig.tv_nsec)
    {
      diff.tv_sec  = t.tv_sec - extrapolate_clock.orig.tv_sec;
      diff.tv_nsec = t.tv_nsec - extrapolate_clock.orig.tv_nsec;
    }
  else
    {
      diff.tv_sec  = t.tv_sec - extrapolate_clock.orig.tv_sec - 1;
      diff.tv_nsec = t.tv_nsec - extrapolate_clock.orig.tv_nsec + 1000000000L;
    }
  double delay = 1000000.0 * diff.tv_sec + diff.tv_nsec / 1000.0;
  return delay;
}

static double extrapolate_clock_translate_time(double t_slave, int n)
{
  return t_slave * extrapolate_clock.nodes[n].skew + extrapolate_clock.nodes[n].offset;
}

static struct nm_clock_method_s extrapolate_clock_method =
  {
   .name           = "extrapolate_clock",
   .init           = &extrapolate_clock_init,
   .get_time_usec  = &extrapolate_clock_get_time_usec,
   .finalize       = NULL,
   .translate_time = &extrapolate_clock_translate_time
  };


/* ********************************************************* */

void do_compute(int seconds)
{
  /* compute deadline */
  struct timespec t;
  clock_gettime(CLOCK_MONOTONIC, &t);
  t.tv_sec += seconds;

  /* matrix multiply */
  #define MATRIX_SIZE 256
  static double matrix1[MATRIX_SIZE][MATRIX_SIZE],
    matrix2[MATRIX_SIZE][MATRIX_SIZE],
    result[MATRIX_SIZE][MATRIX_SIZE];
  static int init_done = 0;
  if(!init_done)
    {
      init_done = 1;
      int i, j;
      for(i = 0; i < MATRIX_SIZE; i++)
        {
          for(j = 0; j < MATRIX_SIZE; j++)
            {
              matrix1[i][j] = 1.0;
              matrix1[i][j] = 2.0;
            }
        }
    }
  for(;;)
    {
      int i, j, k;
      for(i = 0; i < MATRIX_SIZE; i++)
        {
          for(j = 0; j < MATRIX_SIZE; j++)
            {
              result[i][j] = 0;
              for(k = 0; k < MATRIX_SIZE; k++)
                {
                  result[i][j] += matrix1[i][k] * matrix2[k][j];
                }
            }
        }
      struct timespec t2;
      clock_gettime(CLOCK_MONOTONIC, &t2);
      if((t2.tv_sec > t.tv_sec) || ((t2.tv_sec == t.tv_sec) && (t2.tv_nsec > t.tv_nsec)))
        {
          return;
        }
    }
}

/* ********************************************************* */

struct nm_clock_entry_s
{
  double t_master;
  double t_slave;
};

struct nm_clock_drift_s
{
  struct nm_clock_entry_s*p_entries; /**< array indexed by round number */
};

static int rounds = ROUNDS_DEFAULT;

int main(int argc, char**argv)
{
  int compute = 0;
  if(argc == 2 || argc == 3)
    {
      const char*s_rounds = argv[1];
      rounds = atoi(s_rounds);
      if(rounds <= 0)
        {
          fprintf(stderr, "usage: %s [<rounds>]\n", argv[0]);
          abort();
        }
      if(argc == 3)
        {
          fprintf(stderr, "# do computation on slave nodes\n");
          compute = 1;
        }
    }

  MPI_Init(&argc, &argv);
  MPI_Comm_rank(MPI_COMM_WORLD, &rank);
  MPI_Comm_size(MPI_COMM_WORLD, &worldsize);

  int fd = open("/sys/devices/system/clocksource/clocksource0/current_clocksource", O_RDONLY);
  if(fd > 0)
    {
      char buf[256];
      int rc = read(fd, buf, 256);
      buf[255] = '\0';
      if(rc > 0)
        {
          fprintf(stderr, "# node #%d: clock source = %s\n", rank, buf);
        }
      close(fd);
    }

  static struct nm_clock_method_s*methods[] =
    {
      &sync_clocks_method,
      &offset_clock_method,
      &extrapolate_clock_method
    };
  static const int n_methods = sizeof(methods) / sizeof(struct nm_clock_method_s*);

  int m;
  for(m = 0; m < n_methods; m++)
    {
      fprintf(stderr, "# init method '%s'\n", methods[m]->name);
      methods[m]->init();
      methods[m]->p_drift = malloc(worldsize * sizeof(struct nm_clock_drift_s));
      int i;
      for(i = 0; i < worldsize; i++)
        {
          methods[m]->p_drift[i].p_entries = malloc(rounds * sizeof(struct nm_clock_entry_s));
        }
    }

  int round;
  for(round = 0; round < rounds; round++)
    {
      fprintf(stderr, "# ## round = %d / %d; %d seconds\n", round, rounds, ROUNDS_TIME_SECONDS);
      MPI_Barrier(MPI_COMM_WORLD);
      if(rank == 0)
        {
          int dest;
          for(dest = 1; dest < worldsize; dest++)
            {
              /* pingpong to get a common event on both master time and slave time */
              double t_master, t_slave;
              for(m = 0; m < n_methods; m++)
                {
                  generic_sync(dest, methods[m]->get_time_usec, &t_master, &t_slave);
                  methods[m]->p_drift[dest].p_entries[round].t_master = t_master;
                  methods[m]->p_drift[dest].p_entries[round].t_slave = t_slave;
                }
            }
        }
      else
        {
          double t_master, t_slave;
          for(m = 0; m < n_methods; m++)
            {
              generic_sync(0, methods[m]->get_time_usec, &t_master, &t_slave);
            }
        }
      if(compute && (rank != 0) && (round > (rounds / 2)))
        {
          do_compute(ROUNDS_TIME_SECONDS);
        }
      else
        {
          sleep(ROUNDS_TIME_SECONDS);
        }
    }

  /* final sync */
  for(m = 0; m < n_methods; m++)
    {
      if(methods[m]->finalize != NULL)
        {
          (*methods[m]->finalize)();
        }

      if(rank == 0)
        {
          printf("# results for method: '%s'\n", methods[m]->name);
          printf("# round;\t dest;\t t_master;\t t_slave;\t drift to global;\t drift to local \t method\n");
          for(round = 0; round < rounds; round++)
            {
              int i;
              for(i = 1; i < worldsize; i++)
                {
                  const double t_master = methods[m]->p_drift[i].p_entries[round].t_master;
                  const double t_slave_local = methods[m]->p_drift[i].p_entries[round].t_slave;
                  const double t_slave = (*methods[m]->translate_time)(t_slave_local, i);
                  printf("%8d \t %d \t %.3lf \t %.3lf \t %16.3lf \t %16.3lf \t %-16s\n",
                         round, i, t_master, t_slave, t_slave - t_master, t_slave_local - t_master,
                         methods[m]->name);
                  fflush(stdout);
                }
            }
        }
    }

  MPI_Finalize();

  int i;
  for(m = 0; m < n_methods; m++)
    {
      for(i = 0; i < worldsize; i++)
        {
          free(methods[m]->p_drift[i].p_entries);
        }
      free(methods[m]->p_drift);
    }

  return 0;
}
